#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/moduleparam.h>
#include <linux/kallsyms.h>

#include <linux/uaccess.h>
#include <linux/unistd.h>

#include <linux/inet.h>
#include <linux/syscalls.h>
#include <linux/net.h>

/* man pages:
#include <sys/socket.h>
int accept(int sockfd, struct sockaddr *addr, socklen_t *addrlen);
int connect(int sockfd, const struct sockaddr *addr, socklen_t addrlen);
int socketcall(int call, unsigned long *args);
*/

#include "find.c"
#include "CR0.c"

#define PREFIX "! "
#define DEB(...) if( debug ) printk( KERN_INFO PREFIX __VA_ARGS__ )
#define LOG(...) printk( KERN_INFO PREFIX __VA_ARGS__ )
#define ERR(...) printk( KERN_ERR PREFIX __VA_ARGS__ )

static bool debug = false;                      // logical parameter: debug output level
module_param( debug, bool, 0 );
static char* deny;                              // string parameter: denied IP
module_param( deny, charp, 0 );

static void **taddr;                            // table sys_call_table address

static int __init pos_in_table( const char *symbol ) { // position syscall (__NR_*)
   const int last = __NR_process_vm_writev;        // last syscall in i386;
   void *waddr = find_sym( symbol );
   int n;
   if( NULL == waddr ) return -1;
   for( n = 0; n <= last; n++ )
      if( taddr[ n ] == waddr ) break;
   return n <= last ? n : -1;
}

// <linux/syscalls.h> :
asmlinkage long (*old_sys_socketcall) ( int call, unsigned long __user *args );
//asmlinkage long (*old_sys_connect) ( int, struct sockaddr __user *, int );
//asmlinkage long (*old_sys_accept) ( int, struct sockaddr __user *, int __user * );
asmlinkage long new_sys_socketcall( int call, unsigned long __user *args ) {
   long ret;
   if( SYS_CONNECT == call || SYS_ACCEPT == call )
LOG( "new_sys_socketcall before cod=%d\n", call );
   ret = old_sys_socketcall( call, args );
   if( SYS_CONNECT == call || SYS_ACCEPT == call )
LOG( "new_sys_socketcall after cod=%d\n", call );
   // sys_accept & sys_connect` are exported
   //   long aret = sys_accept( 1, (struct sockaddr __user*)args, 0 );
   return ret;
}
EXPORT_SYMBOL( new_sys_socketcall );

static int __init init( void ) {
   // ----------- nested functions are a GCC extension ---------
   void show_in_table( char *symb ) {   // the nested function
      void *waddr = find_sym( symb );
      if( NULL == waddr )
         LOG( "symbol %s not found in kernel\n", symb );
      else {
         int n = pos_in_table( symb );
         if( n > 0 )
            LOG( "symbol %s address = %p, position in sys_call_table = %d\n", symb, waddr, n );
         else
            LOG( "symbol %s address = %p, not found in sys_call_table\n", symb, waddr );
      }
   }
   // --------------------------------------------------------
   void *waddr;
   int ipdeny = deny != NULL ? in_aton( deny ) : in_aton( "127.0.0.1" );
LOG( "deny=%s\n", deny );
LOG( "aton = %d\n", ipdeny );
   if( NULL == ( taddr = find_sym( "sys_call_table" ) ) ) {
      ERR( "sys_call_table not found\n" ); return -EINVAL;
   }
   LOG( "sys_call_table address = %p\n", taddr );
   show_in_table( "sys_accept" );
   show_in_table( "sys_connect" );
   show_in_table( "sys_socketcall" );                       // only diagnostic
   old_sys_socketcall = (void*)taddr[ __NR_socketcall ];
   if( NULL == ( waddr = find_sym( "sys_socketcall" ) ) ) { // sys_socketcall not exported
      ERR( "sys_socketcall not found\n" ); return -EINVAL;
   }
   if( old_sys_socketcall != waddr ) {                      // reinsurance!
      ERR( "Oooops! I don't understand: addresses not equal\n" ); return -EINVAL;
   }
   if( debug ) show_cr0();
   rw_enable();
   taddr[ __NR_socketcall ] = new_sys_socketcall;
   if( debug ) show_cr0();
   rw_disable();
   if( debug ) show_cr0();
   LOG( "install new sys_socketcall handler: %p\n", &new_sys_socketcall );
   return 0;
}

static void __exit exit( void ) {
   LOG( "sys_socketcall handler before unload: %p\n", (void*)taddr[ __NR_socketcall ] );
   rw_enable();
   taddr[ __NR_socketcall ] = old_sys_socketcall;
   rw_disable();
   LOG( "restore old sys_socketcall handler: %p\n", (void*)taddr[ __NR_socketcall ] );
   return;
}

module_init( init );
module_exit( exit );

MODULE_LICENSE( "GPL" );
MODULE_AUTHOR( "Oleg Tsiliuric <olej@front.ru>" );
